package de.hsrm.swt.vokabeltrainer.strict.logik;

import java.util.Vector;

import de.hsrm.swt.vokabeltrainer.lax.data.Vokabel;

/**
 * ber eine Instanz dieser Klasse wird eine Vokabelprfung durchgefhrt: Beim
 * erzeugen wird die Liste der zu prfenden Vokabeln (mind. eine) bergeben.
 * Diese Liste wird in exakt der Listenreihenfolge abgearbeitet.<br/>
 * <br/>
 * Danach wird die Methode {@link Vokabelpruefung#start()} aufgerufen, um die Prfung zu starten.
 * Diese Methode gibt die erste zu prfende Vokabel zurck. Der Verwender zeigt
 * sie dem User an und lt ihn die Antwort eingeben.<br/>
 * <br/>
 * Die eingegene Antwort wird an die Methode {@link Vokabelpruefung#checkVokabel(String)}
 * bergeben. Diese Methode prft, ob die Antwort korrekt ist. Wenn ja, dann
 * gibt sie die nchste Vokabel zurck und zhlt die korrekten Antworten hoch.
 * Wenn nein, dann hat der User bis zu zwei weitere Versuche (also insgesamt
 * drei Fehlversuche). Dies sind jeweils Aufrufe der Methode
 * {@link Vokabelpruefung#checkVokabel(String)} mit der neuen Eingabe. Gibt er bei einem dieser weiteren
 * Versuche die Vokabel korrekt ein, zhlt dies als gltige Vokabel und die
 * nchste zu prfende Vokabel wird zurckgegeben. Gibt der dreimal eine falsche
 * Antwort ein, dann wird dies als Fehlversuch gewertet und es wird ebenfalls
 * die zu prfende Vokabel zurckgegeben. Das Ende der Prfung ist erreicht,
 * wenn  {@link Vokabelpruefung#checkVokabel(String)} {@code NULL} zurckgibt.<br/>
 * <br/>
 * Am Ende kann ber {@link Vokabelpruefung#getStatistikKorrekt} und {@link Vokabelpruefung#getStatistikFalsch} eine
 * Statistik abgerufen werden.<br/>
 * <br/>
 * Definitionen (fr pre/post):<br/>
 * <ul>
 *   <li>Vokabelliste: Liste der aktuell zu prfenden Vokabeln </li>
 *   <li>AktuelleVokabel: dies ist die aktuell zu prfende Vokabel, kann NULL sein, wenn der gesamte Test durchlaufen wurde</li>
 *   <li>StatistikKorrekt: Anzahl korrekt beantworteter Vokabeln</li>
 *   <li>StatistikFalsch: Anzahl falsch beantworteter Vokabeln</li>
 *   <li>AktuelleVokabelFalsch: Anzahl von Fehlversuchen fr aktuelle Vokabel</li>
 * </ul>
 * 
 * @author Wolfgang Knauf
 * 
 */
public class Vokabelpruefung
{
  // Reprsentationsinvariante:
  // listVokabeln.size() >= 1 &&
  // iNextVokabel zwischen 0 und listVokabeln.size()
  // iStatistikKorrekt zwischen 0 und listVokabeln.size() &&
  // iStatistikFalsch zwischen 0 und listVokabeln.size() &&
  // (iStatistikKorrekt + iStatistikFalsch) zwischen 0 und listVokabeln.size() &&
  // iAktuelleVokabelFalsch zwischen 0 und 2

  /**
   * Liste der zu prfenden Vokabeln, wird dem Konstruktor bergeben und enthlt
   * mindestens ein Element.
   */
  private Vector<Vokabel> listVokabeln;

  /**
   * Index der aktuell abgeprften Vokabel. Muss per {@link Vokabelpruefung#start} initialisiert
   * werden, bei jeder richtigen oder bei jeder (nach drei Fehlversuchen) falsch
   * beantworteten Vokabelfrage wird der Index um 1 hochgezhlt und steht am
   * Ende des Tests auf "{@link Vokabelpruefung#listVokabeln}.size() + 1"
   */
  private int iNextVokabel = 0;

  /**
   * Anzahl der korrekt beantworteten Vokabeln, wird in {@link Vokabelpruefung#checkVokabel} bei
   * korrekter Antwort hochgezhlt.
   */
  private int iStatistikKorrekt = 0;

  /**
   * Anzahl der mit drei Fehlversuchen nicht beantworteten Vokabeln, wird in 
   * {@link Vokabelpruefung#checkVokabel} bei korrekter Antwort hochgezhlt.
   */
  private int iStatistikFalsch = 0;

  /**
   * Falls eine Vokabel das erste Mal falsch beantwortet wurde, wird diese
   * Variable auf 1 gesetzt. Dito bei der zweiten falschen Antwort. Wird die
   * Vokabel erneut falsch beantwortet, dann wird die Variable wieder auf "0"
   * gesetzt. Dito bei korrekter Antwort nach ein oder zwei Fehlversuchen.
   * 
   * Anhand dieser Variable erkennt die Methode {@link Vokabelpruefung#checkVokabel(String)}, ob
   * eine Vokabel dreimal falsch beantwortet wurde.
   */
  private int iAktuelleVokabelFalsch = 0;

  /**
   * Anzahl der korrekt beantworteten Vokabeln, wird in {@link Vokabelpruefung#checkVokabel} bei
   * korrekter Antwort hochgezhlt. Abrufen der Statistik ist jederzeit whrend
   * (oder vor dem Vokabeltest) mglich.
   * 
   * @return Anzahl korrekt beantworteter Vokabeln.
   */
  public int getStatistikKorrekt()
  {
    return iStatistikKorrekt;
  }

  /**
   * Anzahl der mit drei Fehlversuchen nicht beantworteten Vokabeln, wird in
   * {@link Vokabelpruefung#checkVokabel} bei korrekter Antwort hochgezhlt. Abrufen der Statistik ist
   * jederzeit whrend (oder vor dem Vokabeltest) mglich.
   * 
   * @return Anzahl falsch beantworteter Vokabeln.
   */
  public int getStatistikFalsch()
  {
    return iStatistikFalsch;
  }

  
  /**
   * Anzahl der Fehlversuche im aktuellen Wort: wird bei jeder an
   * {@link Vokabelpruefung#checkVokabel} bergebenen falschen Antwort hochgezhlt.
   * Bei einer richtigen Antwort oder bei der dritten falschen Antwort wird
   * der Zhler auf 0 gestellt.  
   * Abrufen dieses Werts ist jederzeit whrend (oder vor dem Vokabeltest) mglich.
   * 
   * @return Anzahl falscher Antworten bei aktueller Vokabel.
   */
  public int getAktuelleVokabelFalsch()
  {
    return this.iAktuelleVokabelFalsch;
  }


  /**
   * Initialisierung der Vokabelprfung: es wird die Liste von zu prfenden
   * Vokabeln bergeben.
   * 
   * @post this = () && Vokabelliste == listVokabeln
   * @param listVokabeln
   *          Liste der zu prfenden Vokabeln. Muss mindestens ein Element
   *          enthalten und darf nicht NULL sein.
   * @throws NullPointerException
   *           Wenn listVokabeln = null
   * @throws IllegalArgumentException
   *           Wenn listVokabeln nur ein Element enthlt.
   */
  public Vokabelpruefung(Vector<Vokabel> listVokabeln)
  {
    if (listVokabeln == null)
    {
      throw new NullPointerException("listVokabeln darf nicht NULL sein");
    }
    if (listVokabeln.size() == 0)
    {
      throw new IllegalArgumentException("listVokabeln darf nicht leer sein");
    }

    this.listVokabeln = listVokabeln;

    // Reprsentationsinvariante zur Sicherheit auch hier prfen:
    this.checkRep();
  }

  /**
   * Starten der Vokabelprfung: liefert die erste zu prfende Vokabel (= erste
   * Vokabel in der per Konstruktor bergebenen Liste) zurck. Da sie alle
   * Zhlvariablen zurcksetzt, kann sie auch nach einer durchlaufenen Prfung
   * oder mittendrin als "Neustart" aufgerufen werden.
   * 
   * @post AktuelleVokabel == erstes Element aus Vokabelliste && StatistikKorrekt == 0 &&
   *       StatistikFalsch == 0 && AktuelleVokabelFalsch == 0
   * 
   * @return Erste zu prfende Vokabel.
   */
  public Vokabel start()
  {
    // Reprsentationsinvariante vorher prfen:
    this.checkRep();

    // Alle Zhler zurcksetzen:
    this.iNextVokabel = 0;

    this.iStatistikFalsch = 0;
    this.iStatistikKorrekt = 0;
    this.iAktuelleVokabelFalsch = 0;

    // Erste zu prfende Vokabel zurckgeben.
    try
    {
      return this.listVokabeln.get(this.iNextVokabel);
    }
    finally
    {
      // Reprsentationsinvariante danach prfen:
      this.checkRep();
    }
  }

  /**
   * Prft eine Vokabel darauf, ob die korrekte Anwort bergeben wurde. Rckgabe
   * der Methode ist die nchste zu prfende Vokabel oder im Falle einer
   * falschen Antwort die aktuelle Vokabel. Nach drei Fehlversuchen kommt
   * automatisch die nchste Vokabel. Hat man die letzte Vokabel der Liste
   * korrekt oder dreimal falsch beantwortet, gibt sie NULL zurck.<br/>
   * <br/>
   * Beim Prfen auf Korrektheit wird automatisch die Statistik angepat:
   * <ul>
   *   <li>korrekte Antwort: erhhe Zhler fr korrekte Antworten</li>
   *   <li>erste oder zweite falsche Antwort: tue nichts mit Statistik</li>
   *   <li>dritte falsche Antwort: erhhe Zhler fr falsche Antworten.</li>
   * </ul>
   * 
   * @param sAntwort
   *          Antwort-Versuch, darf auch NULL oder leer sein (das wre immer
   *          eine falsche Antwort).
   * @return bei korrekter oder dritter falscher Antwort: nchste zu prfende
   *         Vokabel oder NULL bei erreichen der Liste. bei erster und zweiter
   *         falscher Antwort: aktuelle Vokabel Bei korrekter oder falscher
   *         Beantwortung der letzten Vokabel der Liste: NULL
   * 
   * @throws IllegalStateException
   *           Falls Vokabelliste bereits komplett durchlaufen wurde.
   * 
   * @pre AktuelleVokabel != null (es gibt noch unbeantwortete Vokabeln)
   * @post <ul>
   *         <li>Wenn sAntwort == AktuelleVokabel.Antwort: <br/>
   *               AktuelleVokabel = Vokabelliste + 1 && StatistikKorrekt = StatistikKorrekt + 1 && AktuelleVokabelFalsch == 0</li> 
   *         <li>Wenn sAntwort != AktuelleVokabel.Antwort && (AktuelleVokabelFalsch == 0 || AktuelleVokabelFalsch == 1):<br/> 
   *               Aktuelle Vokabel = AktuelleVokabel && AktuelleVokabelFalsch == AktuelleVokabelFalsch + 1</li> 
   *         <li>Wenn sAntwort != AktuelleVokabel.Antwort && (AktuelleVokabelFalsch == 2):<br/> 
   *               Aktuelle Vokabel = Vokabelliste + 1 && StatistikFalsch = StatistikFalsch + 1 AktuelleVokabelFalsch == 0</li>
   *       </ul> 
   */
  public Vokabel checkVokabel(String sAntwort)
  {
    // Reprsentationsinvariante vorher prfen:
    this.checkRep();

    // Man darf nicht ber das Ende der Vokabeln hinauslaufen:
    if (this.iNextVokabel >= this.listVokabeln.size())
    {
      throw new IllegalStateException("Vokabelliste wurde bereits komplett durchlaufen");
    }

    // Folgendes in try/finally-Block, damit die Reprsentationsinvariante bei
    // JEDER der mglichen Rckgabepfade
    // geprft wird.
    try
    {
      if (this.listVokabeln.get(this.iNextVokabel).getAntwort().equals(sAntwort) == true)
      {
        // Korrekte Antwort:
        // Statistik der richtigen hochzhlen.
        ++this.iStatistikKorrekt;

        // Falls vorher Fehlversuche passierten, dann den Zhler dafr
        // zurcksetzen.
        this.iAktuelleVokabelFalsch = 0;

        // Zur nchsten Vokabel springen. Falls man dabei ans Ende der Liste
        // kommt, dann kommt NULL zurck.
        ++this.iNextVokabel;
        if (this.iNextVokabel < this.listVokabeln.size())
        {
          return this.listVokabeln.get(this.iNextVokabel);
        }
        else
        {
          return null;
        }
      }
      else
      {
        // Falsche Antwort:
        // Ist das der erste oder zweite Fehlversuch? Wenn ja, dann geht es mit
        // der gleichen Vokabel weiter.
        if (this.iAktuelleVokabelFalsch < 2)
        {
          // Fehlversuch-Zhler hochzhlen.
          ++this.iAktuelleVokabelFalsch;
          return this.listVokabeln.get(this.iNextVokabel);
        }
        else
        {
          // dreimal falsch: nchste!
          this.iAktuelleVokabelFalsch = 0;

          // Statistik der falschen anpassen.
          ++this.iStatistikFalsch;

          // Zur nchsten Vokabel springen. Falls man dabei ans Ende der Liste
          // kommt, dann kommt NULL zurck.
          ++this.iNextVokabel;
          if (this.iNextVokabel < this.listVokabeln.size())
          {
            return this.listVokabeln.get(this.iNextVokabel);
          }
          else
          {
            return null;
          }
        }
      }
    }
    finally
    {
      // Reprsentationsinvariante vorher prfen:
      this.checkRep();
    }
  }

  /**
   * Prft die Reprsentationsinvariante - siehe Klassenkommentar. Wirft
   * Exceptions, wenn irgendetwas bse ist. Ansonsten macht die Methode nichts.
   */
  private void checkRep()
  {
    if (this.listVokabeln == null)
    {
      // Kann eigentlich nicht passieren ;-)
      throw new IllegalStateException("listvokabeln darf nicht NULL sein");
    }
    if (this.listVokabeln.size() == 0)
    {
      // Das knnte passieren, wenn die per Konstruktor bergebene Liste
      // auerhalb manipuliert wird (Items werden entfernt),
      // da sie beim Zuweisen der Membervariablen nicht kopiert wird.
      throw new IllegalStateException("listvokabeln muss mind. ein Element enthalten");
    }

    if (this.iNextVokabel < 0 || this.iNextVokabel > this.listVokabeln.size())
    {
      throw new IllegalStateException("iNextVokabel ungltig: " + this.iNextVokabel + " - muss zwischen 0 und "
          + this.listVokabeln.size() + " liegen");
    }
    if (this.iStatistikKorrekt < 0 || this.iStatistikKorrekt > this.listVokabeln.size())
    {
      throw new IllegalStateException("iStatistikKorrekt ungltig: " + this.iStatistikKorrekt
          + " - muss zwischen 0 und " + this.listVokabeln.size() + " liegen");
    }
    if (this.iStatistikFalsch < 0 || this.iStatistikFalsch > this.listVokabeln.size())
    {
      throw new IllegalStateException("iStatistikFalsch ungltig: " + this.iStatistikFalsch + " - muss zwischen 0 und "
          + this.listVokabeln.size() + " liegen");
    }
    if (this.iAktuelleVokabelFalsch < 0 || this.iAktuelleVokabelFalsch > 2)
    {
      throw new IllegalStateException("iAktuelleVokabelFalsch ungltig: " + this.iAktuelleVokabelFalsch
          + " - muss zwischen 0 und 2 liegen");
    }
  }
  
  /**Gibt die gesamte externe Reprsentation der Klasse als String aus (Abstraktionsfunktion)
   * Da hier alle Membervariablen der Klasse nach auen abrufbar sind
   * (bzw. ber die von auen gesetzte Vokabelliste auerhalb bekannt sein sollte)
   * entspricht dies einer Ausgabe der Reprsentationsinvariante. 
   * 
   * @return Externe Reprsentation der Klasse mit allen Informationen
   */
  @Override
  public String toString()
  {
    String sNewline = System.getProperty("line.separator");
    return "Anzahl Vokabeln: " + this.listVokabeln.size() + sNewline + 
        "Index der aktuellen Vokabel: " + this.iNextVokabel + sNewline + 
        "Aktuelle Vokabel: " + (this.iNextVokabel < this.listVokabeln.size() ? this.listVokabeln.get(this.iNextVokabel).toString() : "(null)") + sNewline +
        "Statistik: korrekt: " + this.iStatistikKorrekt + sNewline + 
        "Statistik: falsch: " + this.iStatistikFalsch + sNewline + 
        "Aktuelle Fehlversuche: " + this.iAktuelleVokabelFalsch; 
        
  }

}
